"""
For ease of lecture we split DP implementations from standard implementations.

However, the DP op need to be defined before the standard op since the standard
op holds a reference to the DP op.

isort:skip_file
"""
import logging

logger = logging.getLogger(__name__)

try:
    from .pandas_dp import *  # noqa: F403
except NameError:
    logger.info("DP pandas ops not available.")

from .pandas import *  # noqa: F403,E402



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/dataset.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/dataset.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&sarus_data_spec/protobuf/dataset.proto\x12\x0fsarus_data_spec\"\xa3\x07\n\x07\x44\x61taset\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12+\n\x04spec\x18\x04 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.Spec\x12<\n\nproperties\x18\x05 \x03(\x0b\x32(.sarus_data_spec.Dataset.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x8d\x02\n\x04Spec\x12;\n\x0btransformed\x18\x01 \x01(\x0b\x32$.sarus_data_spec.Dataset.TransformedH\x00\x12-\n\x04\x66ile\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Dataset.FileH\x00\x12/\n\x05\x66iles\x18\x03 \x01(\x0b\x32\x1e.sarus_data_spec.Dataset.FilesH\x00\x12\x33\n\x07\x61rchive\x18\x04 \x01(\x0b\x32 .sarus_data_spec.Dataset.ArchiveH\x00\x12+\n\x03sql\x18\x05 \x01(\x0b\x32\x1c.sarus_data_spec.Dataset.SqlH\x00\x42\x06\n\x04spec\x1a\xbd\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12Q\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x38.sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a#\n\x04\x46ile\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03uri\x18\x02 \x01(\t\x1a,\n\x05\x46iles\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x13\n\x0buri_pattern\x18\x02 \x01(\t\x1a>\n\x07\x41rchive\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x16\n\x0e\x61rchive_format\x18\x02 \x01(\t\x12\x0b\n\x03uri\x18\x03 \x01(\t\x1an\n\x03Sql\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x32\n\x06tables\x18\x02 \x03(\x0b\x32\".sarus_data_spec.Dataset.Sql.Table\x1a&\n\x05Table\x12\x0e\n\x06schema\x18\x01 \x01(\t\x12\r\n\x05table\x18\x02 \x01(\tb\x06proto3'
)




_DATASET_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Dataset.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Dataset.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Dataset.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=219,
  serialized_end=268,
)

_DATASET_SPEC = _descriptor.Descriptor(
  name='Spec',
  full_name='sarus_data_spec.Dataset.Spec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformed', full_name='sarus_data_spec.Dataset.Spec.transformed', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file', full_name='sarus_data_spec.Dataset.Spec.file', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='files', full_name='sarus_data_spec.Dataset.Spec.files', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='archive', full_name='sarus_data_spec.Dataset.Spec.archive', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sql', full_name='sarus_data_spec.Dataset.Spec.sql', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='spec', full_name='sarus_data_spec.Dataset.Spec.spec',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=271,
  serialized_end=540,
)

_DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY = _descriptor.Descriptor(
  name='NamedArgumentsEntry',
  full_name='sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=679,
  serialized_end=732,
)

_DATASET_TRANSFORMED = _descriptor.Descriptor(
  name='Transformed',
  full_name='sarus_data_spec.Dataset.Transformed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transform', full_name='sarus_data_spec.Dataset.Transformed.transform', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='sarus_data_spec.Dataset.Transformed.arguments', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='named_arguments', full_name='sarus_data_spec.Dataset.Transformed.named_arguments', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=543,
  serialized_end=732,
)

_DATASET_FILE = _descriptor.Descriptor(
  name='File',
  full_name='sarus_data_spec.Dataset.File',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format', full_name='sarus_data_spec.Dataset.File.format', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uri', full_name='sarus_data_spec.Dataset.File.uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=734,
  serialized_end=769,
)

_DATASET_FILES = _descriptor.Descriptor(
  name='Files',
  full_name='sarus_data_spec.Dataset.Files',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format', full_name='sarus_data_spec.Dataset.Files.format', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uri_pattern', full_name='sarus_data_spec.Dataset.Files.uri_pattern', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=771,
  serialized_end=815,
)

_DATASET_ARCHIVE = _descriptor.Descriptor(
  name='Archive',
  full_name='sarus_data_spec.Dataset.Archive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format', full_name='sarus_data_spec.Dataset.Archive.format', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='archive_format', full_name='sarus_data_spec.Dataset.Archive.archive_format', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uri', full_name='sarus_data_spec.Dataset.Archive.uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=817,
  serialized_end=879,
)

_DATASET_SQL_TABLE = _descriptor.Descriptor(
  name='Table',
  full_name='sarus_data_spec.Dataset.Sql.Table',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='sarus_data_spec.Dataset.Sql.Table.schema', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='sarus_data_spec.Dataset.Sql.Table.table', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=953,
  serialized_end=991,
)

_DATASET_SQL = _descriptor.Descriptor(
  name='Sql',
  full_name='sarus_data_spec.Dataset.Sql',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='sarus_data_spec.Dataset.Sql.uri', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tables', full_name='sarus_data_spec.Dataset.Sql.tables', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASET_SQL_TABLE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=881,
  serialized_end=991,
)

_DATASET = _descriptor.Descriptor(
  name='Dataset',
  full_name='sarus_data_spec.Dataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='sarus_data_spec.Dataset.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Dataset.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='doc', full_name='sarus_data_spec.Dataset.doc', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spec', full_name='sarus_data_spec.Dataset.spec', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Dataset.properties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASET_PROPERTIESENTRY, _DATASET_SPEC, _DATASET_TRANSFORMED, _DATASET_FILE, _DATASET_FILES, _DATASET_ARCHIVE, _DATASET_SQL, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=991,
)

_DATASET_PROPERTIESENTRY.containing_type = _DATASET
_DATASET_SPEC.fields_by_name['transformed'].message_type = _DATASET_TRANSFORMED
_DATASET_SPEC.fields_by_name['file'].message_type = _DATASET_FILE
_DATASET_SPEC.fields_by_name['files'].message_type = _DATASET_FILES
_DATASET_SPEC.fields_by_name['archive'].message_type = _DATASET_ARCHIVE
_DATASET_SPEC.fields_by_name['sql'].message_type = _DATASET_SQL
_DATASET_SPEC.containing_type = _DATASET
_DATASET_SPEC.oneofs_by_name['spec'].fields.append(
  _DATASET_SPEC.fields_by_name['transformed'])
_DATASET_SPEC.fields_by_name['transformed'].containing_oneof = _DATASET_SPEC.oneofs_by_name['spec']
_DATASET_SPEC.oneofs_by_name['spec'].fields.append(
  _DATASET_SPEC.fields_by_name['file'])
_DATASET_SPEC.fields_by_name['file'].containing_oneof = _DATASET_SPEC.oneofs_by_name['spec']
_DATASET_SPEC.oneofs_by_name['spec'].fields.append(
  _DATASET_SPEC.fields_by_name['files'])
_DATASET_SPEC.fields_by_name['files'].containing_oneof = _DATASET_SPEC.oneofs_by_name['spec']
_DATASET_SPEC.oneofs_by_name['spec'].fields.append(
  _DATASET_SPEC.fields_by_name['archive'])
_DATASET_SPEC.fields_by_name['archive'].containing_oneof = _DATASET_SPEC.oneofs_by_name['spec']
_DATASET_SPEC.oneofs_by_name['spec'].fields.append(
  _DATASET_SPEC.fields_by_name['sql'])
_DATASET_SPEC.fields_by_name['sql'].containing_oneof = _DATASET_SPEC.oneofs_by_name['spec']
_DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY.containing_type = _DATASET_TRANSFORMED
_DATASET_TRANSFORMED.fields_by_name['named_arguments'].message_type = _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY
_DATASET_TRANSFORMED.containing_type = _DATASET
_DATASET_FILE.containing_type = _DATASET
_DATASET_FILES.containing_type = _DATASET
_DATASET_ARCHIVE.containing_type = _DATASET
_DATASET_SQL_TABLE.containing_type = _DATASET_SQL
_DATASET_SQL.fields_by_name['tables'].message_type = _DATASET_SQL_TABLE
_DATASET_SQL.containing_type = _DATASET
_DATASET.fields_by_name['spec'].message_type = _DATASET_SPEC
_DATASET.fields_by_name['properties'].message_type = _DATASET_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['Dataset'] = _DATASET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASET_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.PropertiesEntry)
    })
  ,

  'Spec' : _reflection.GeneratedProtocolMessageType('Spec', (_message.Message,), {
    'DESCRIPTOR' : _DATASET_SPEC,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Spec)
    })
  ,

  'Transformed' : _reflection.GeneratedProtocolMessageType('Transformed', (_message.Message,), {

    'NamedArgumentsEntry' : _reflection.GeneratedProtocolMessageType('NamedArgumentsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY,
      '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Transformed.NamedArgumentsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASET_TRANSFORMED,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Transformed)
    })
  ,

  'File' : _reflection.GeneratedProtocolMessageType('File', (_message.Message,), {
    'DESCRIPTOR' : _DATASET_FILE,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.File)
    })
  ,

  'Files' : _reflection.GeneratedProtocolMessageType('Files', (_message.Message,), {
    'DESCRIPTOR' : _DATASET_FILES,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Files)
    })
  ,

  'Archive' : _reflection.GeneratedProtocolMessageType('Archive', (_message.Message,), {
    'DESCRIPTOR' : _DATASET_ARCHIVE,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Archive)
    })
  ,

  'Sql' : _reflection.GeneratedProtocolMessageType('Sql', (_message.Message,), {

    'Table' : _reflection.GeneratedProtocolMessageType('Table', (_message.Message,), {
      'DESCRIPTOR' : _DATASET_SQL_TABLE,
      '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Sql.Table)
      })
    ,
    'DESCRIPTOR' : _DATASET_SQL,
    '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset.Sql)
    })
  ,
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'sarus_data_spec.protobuf.dataset_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Dataset)
  })
_sym_db.RegisterMessage(Dataset)
_sym_db.RegisterMessage(Dataset.PropertiesEntry)
_sym_db.RegisterMessage(Dataset.Spec)
_sym_db.RegisterMessage(Dataset.Transformed)
_sym_db.RegisterMessage(Dataset.Transformed.NamedArgumentsEntry)
_sym_db.RegisterMessage(Dataset.File)
_sym_db.RegisterMessage(Dataset.Files)
_sym_db.RegisterMessage(Dataset.Archive)
_sym_db.RegisterMessage(Dataset.Sql)
_sym_db.RegisterMessage(Dataset.Sql.Table)


_DATASET_PROPERTIESENTRY._options = None
_DATASET_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import typing as t

import pyarrow as pa

from sarus_data_spec.manager.asyncio.utils import decoupled_async_iter
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.typing as st


class DataspecStaticChecker:
    def __init__(self, dataspec: st.DataSpec):
        self.dataspec = dataspec

    def is_dp_applicable(self, public_context: t.Collection[str]) -> bool:
        """Statically check if a DP transform is applicable in this position.

        This verification is common to all dataspecs and is true if:
            - the dataspec is transformed and its transform has an equivalent
            DP transform
            - the DP transform's required PEP arguments are PEP and aligned
            (i.e. same PEP token)
            - other dataspecs arguments are public
        """
        return False

    def is_dp(self) -> bool:
        """Checks if the transform is DP and compatible with the arguments."""
        return False

    def dp_transform(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        return None

    def dp_equivalent(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        raise NotImplementedError

    async def private_queries(self) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""
        if self.is_dp():
            raise NotImplementedError
        else:
            return []


class DatasetStaticChecker(DataspecStaticChecker):
    def __init__(self, dataset: st.Dataset):
        super().__init__(dataset)
        self.dataset = dataset

    async def schema(self) -> st.Schema:
        """Computes the schema of the dataspec"""
        raise NotImplementedError

    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """Return a token if the output is PEP."""
        raise NotImplementedError

    def pep_kind(self) -> sdvt.PEPKind:
        raise NotImplementedError


class DatasetImplementation:
    def __init__(self, dataset: st.Dataset):
        self.dataset = dataset

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError

    async def size(self) -> st.Size:
        raise NotImplementedError

    async def bounds(self) -> st.Bounds:
        raise NotImplementedError

    async def marginals(self) -> st.Marginals:
        raise NotImplementedError

    async def sql(
        self,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """It composes the query and it sends it to the parent."""
        raise NotImplementedError

    @staticmethod
    async def decoupled_async_iter(
        source: t.AsyncIterator[pa.RecordBatch], buffer_size: int = 100
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return decoupled_async_iter(source=source, buffer_size=buffer_size)


class ScalarImplementation:
    def __init__(self, scalar: st.Scalar):
        self.scalar = scalar

    async def value(self) -> t.Any:
        raise NotImplementedError

    @staticmethod
    async def decoupled_async_iter(
        source: t.AsyncIterator[pa.RecordBatch], buffer_size: int = 100
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return decoupled_async_iter(source=source, buffer_size=buffer_size)



"""Protocols describing common object behaviors
"""
from typing import (
    ByteString,
    MutableMapping,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from google.protobuf.descriptor import Descriptor

M = TypeVar('M')


@runtime_checkable
class Protobuf(Protocol):
    """A protobuf message.
    Only the used methods are defined in this protocol.
    We assume all protobuf messages have properties
    """

    DESCRIPTOR: Descriptor

    def CopyFrom(self: M, other_msg: M) -> None:
        ...

    def SerializeToString(self: M, deterministic: bool) -> bytes:
        ...

    def FromString(self: M, s: Union[ByteString, bytes]) -> M:
        ...

    @property
    def properties(self) -> MutableMapping:
        ...


@runtime_checkable
class ProtobufWithUUID(Protobuf, Protocol):
    uuid: str


@runtime_checkable
class ProtobufWithUUIDAndDatetime(ProtobufWithUUID, Protocol):
    datetime: str  # ISO 8601 datetime



import typing as t

from sarus_data_spec.protobuf.attribute_pb2 import Attribute
from sarus_data_spec.protobuf.bounds_pb2 import Bounds
from sarus_data_spec.protobuf.constraint_pb2 import (
    ConstraintKind,
    VariantConstraint,
)
from sarus_data_spec.protobuf.dataset_pb2 import Dataset
from sarus_data_spec.protobuf.links_pb2 import Links
from sarus_data_spec.protobuf.manager_pb2 import Manager
from sarus_data_spec.protobuf.marginals_pb2 import Marginals
from sarus_data_spec.protobuf.path_pb2 import Path
from sarus_data_spec.protobuf.predicate_pb2 import Predicate
from sarus_data_spec.protobuf.proto_container_pb2 import ProtoContainer
from sarus_data_spec.protobuf.relation_pb2 import Relation
from sarus_data_spec.protobuf.scalar_pb2 import Scalar
from sarus_data_spec.protobuf.schema_pb2 import Schema
from sarus_data_spec.protobuf.size_pb2 import Size
from sarus_data_spec.protobuf.statistics_pb2 import Distribution, Statistics
from sarus_data_spec.protobuf.status_pb2 import Status
from sarus_data_spec.protobuf.transform_pb2 import Transform
from sarus_data_spec.protobuf.type_pb2 import Type
from sarus_data_spec.protobuf.utilities import (
    copy,
    dejson,
    deserialize,
    dict_deserialize,
    dict_serialize,
    from_base64,
    json,
    json_deserialize,
    json_serialize,
    message,
    message_type,
    serialize,
    to_base64,
    type_name,
    unwrap,
    wrap,
)


def python_proto_factory(string_to_deserialize: str, name: str) -> t.Any:
    proto_type = message(type_name=name)
    proto = from_base64(string_to_deserialize, proto_type)

    if name == type_name(Path):
        import sarus_data_spec.path as path_module

        proto = t.cast(Path, proto)
        return path_module.Path(proto)

    elif name == type_name(Type):
        import sarus_data_spec.type as type_module

        proto = t.cast(Type, proto)
        return type_module.Type(proto)

    raise NotImplementedError(f'current {name} not implemented yet')



from __future__ import annotations

from typing import Dict, Type

from sarus_data_spec.base import Referring
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Attribute(Referring[sp.Attribute]):
    """A python class to describe attributes"""

    def __init__(self, protobuf: sp.Attribute, store: bool = True) -> None:
        self._referred = {protobuf.object}
        # This has to be defined before it is initialized
        super().__init__(protobuf=protobuf, store=store)

    def prototype(self) -> Type[sp.Attribute]:
        """Return the type of the underlying protobuf."""
        return sp.Attribute

    def name(self) -> str:
        return self.protobuf().name


def attach_properties(
    dataspec: st.DataSpec, name: str = "", properties: Dict[str, str] = {}
) -> Attribute:
    return Attribute(
        sp.Attribute(object=dataspec.uuid(), properties=properties, name=name)
    )



from time import time_ns
from typing import Collection, Dict, List, Optional, Tuple, cast
import logging

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.constants import VARIANT_UUID
from sarus_data_spec.context import global_context
from sarus_data_spec.dataset import transformed
from sarus_data_spec.dataspec_validator.typing import DataspecValidator
from sarus_data_spec.manager.ops.asyncio.processor import routing
from sarus_data_spec.scalar import privacy_budget
from sarus_data_spec.variant_constraint import (
    dp_constraint,
    mock_constraint,
    syn_constraint,
)
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

ArgStruct = Tuple[List[int], List[str]]
logger = logging.getLogger(__name__)


def flatten_args(
    args: List[st.DataSpec], kwargs: Dict[str, st.DataSpec]
) -> Tuple[List[st.Dataset], List[st.Scalar], ArgStruct]:
    """Split args and kwargs into Datasets and Scalars."""
    flat_args = args + list(kwargs.values())
    ds_args = [
        cast(st.Dataset, arg)
        for arg in flat_args
        if arg.prototype() == sp.Dataset
    ]
    sc_args = [
        cast(st.Scalar, arg)
        for arg in flat_args
        if arg.prototype() == sp.Scalar
    ]

    ds_idx = [
        i for i, arg in enumerate(flat_args) if arg.prototype() == sp.Dataset
    ]
    sc_idx = [
        i for i, arg in enumerate(flat_args) if arg.prototype() == sp.Scalar
    ]
    idx = ds_idx + sc_idx
    struct = (idx, list(kwargs.keys()))

    return ds_args, sc_args, struct


def nest_args(
    ds_args: List[st.DataSpec],
    sc_args: List[st.DataSpec],
    struct: ArgStruct,
) -> Tuple[List[st.DataSpec], Dict[str, st.DataSpec]]:
    """Nest Datasets and Scalars into args and kwargs."""
    idx, keys = struct
    all_args = ds_args + sc_args
    flat_args = [all_args[idx.index(i)] for i in range(len(idx))]
    n_args = len(flat_args) - len(keys)
    args = flat_args[:n_args]
    kwargs = {key: val for key, val in zip(keys, flat_args[n_args:])}
    return args, kwargs


def attach_variant(
    original: st.DataSpec,
    variant: st.DataSpec,
    kind: st.ConstraintKind,
) -> None:
    attach_properties(
        original,
        properties={
            # TODO deprecated in SDS >= 2.0.0 -> use only VARIANT_UUID
            kind.name: variant.uuid(),
            VARIANT_UUID: variant.uuid(),
        },
        name=kind.name,
    )


def compile(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    privacy_limit: Optional[st.PrivacyLimit],
) -> Optional[st.DataSpec]:
    """Returns a compliant Node or None."""

    if kind == st.ConstraintKind.SYNTHETIC:
        variant, _ = compile_synthetic(
            dataspec_validator,
            dataspec,
            public_context,
        )
        return variant

    elif kind == st.ConstraintKind.MOCK:
        mock_variant, _ = compile_mock(
            dataspec_validator,
            dataspec,
            public_context,
        )
        return mock_variant

    if privacy_limit is None:
        raise ValueError(
            "Privacy limit must be defined for PEP or DP compilation"
        )

    if kind == st.ConstraintKind.DP:
        variant, _ = compile_dp(
            dataspec_validator,
            dataspec,
            public_context=public_context,
            privacy_limit=privacy_limit,
        )
        return variant

    elif kind == st.ConstraintKind.PEP:
        raise NotImplementedError("PEP compilation")

    else:
        raise ValueError(
            f"Privacy policy {kind} compilation not implemented yet"
        )


def compile_synthetic(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
) -> Tuple[st.DataSpec, Collection[str]]:
    # Current dataspec verifies the constraint?
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            constraint,
            st.ConstraintKind.SYNTHETIC,
            public_context,
            privacy_limit=None,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        if variant is None:
            logger.info(f"Found a None variant for dataspec {dataspec.uuid()}")
            continue
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                constraint,
                st.ConstraintKind.SYNTHETIC,
                public_context,
                privacy_limit=None,
            ):
                return variant, public_context

    # Derive the SD from the parents SD
    if dataspec.is_transformed():
        transform = dataspec.transform()
        args, kwargs = dataspec.parents()
        ds_args, sc_args, struct = flatten_args(args, kwargs)
        ds_syn_args_context = [
            compile_synthetic(dataspec_validator, parent, public_context)
            for parent in ds_args
        ]
        sc_syn_args_context = [
            compile_synthetic(dataspec_validator, parent, public_context)
            for parent in sc_args
        ]

        if len(ds_syn_args_context) > 0:
            ds_syn_args, ds_contexts = zip(*ds_syn_args_context)
        else:
            ds_syn_args, ds_contexts = [], ([],)
        if len(sc_syn_args_context) > 0:
            sc_syn_args, sc_contexts = zip(*sc_syn_args_context)
        else:
            sc_syn_args, sc_contexts = [], ([],)
        new_context = cast(
            Collection[str],
            list(set(sum(map(list, ds_contexts + sc_contexts), []))),
        )
        args, kwargs = nest_args(
            cast(List[st.DataSpec], list(ds_syn_args)),
            cast(List[st.DataSpec], list(sc_syn_args)),
            struct,
        )
        syn_variant = cast(
            st.DataSpec,
            transformed(
                transform,
                *args,
                dataspec_type=sp.type_name(dataspec.prototype()),
                dataspec_name=None,
                **kwargs,
            ),
        )
        syn_constraint(
            dataspec=syn_variant, required_context=list(public_context)
        )
        attach_variant(dataspec, syn_variant, kind=st.ConstraintKind.SYNTHETIC)
        return syn_variant, new_context

    elif dataspec.is_public():
        return dataspec, public_context
    else:
        raise TypeError(
            'Non public source Datasets cannot'
            'be compiled to Synthetic, a synthetic variant'
            'should have been created downstream in the graph.'
        )


def compile_mock(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
) -> Tuple[Optional[st.DataSpec], Collection[str]]:
    """Compile the MOCK variant of a DataSpec.

    Note that the MOCK compilation only makes sense for internally transformed
    dataspecs. For externally transformed dataspecs, the MOCK is computed
    before the dataspec, so we can only fetch it.
    """
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            constraint,
            st.ConstraintKind.MOCK,
            public_context,
            privacy_limit=None,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        if variant is None:
            logger.info(f"Found a None variant for dataspec {dataspec.uuid()}")
            continue
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                constraint,
                st.ConstraintKind.MOCK,
                public_context,
                privacy_limit=None,
            ):
                return variant, public_context

    if dataspec.is_public():
        return dataspec, public_context

    if not dataspec.is_transformed():
        raise ValueError(
            'Cannot compile the MOCK of a non public source DataSpec. '
            'A MOCK should be set manually downstream in the '
            'computation graph.'
        )

    # The DataSpec is the result of an internal transform
    transform = dataspec.transform()
    args, kwargs = dataspec.parents()
    mock_args = [arg.variant(st.ConstraintKind.MOCK) for arg in args]
    named_mock_args = {
        name: arg.variant(st.ConstraintKind.MOCK)
        for name, arg in kwargs.items()
    }
    if any([m is None for m in mock_args]) or any(
        [m is None for m in named_mock_args.values()]
    ):
        raise ValueError(
            f"Cannot derive a mock for {dataspec} "
            "because of of the parent has a None MOCK."
        )

    typed_mock_args = [cast(st.DataSpec, ds) for ds in mock_args]
    typed_named_mock_args = {
        name: cast(st.DataSpec, ds) for name, ds in named_mock_args.items()
    }

    mock: st.DataSpec = transformed(
        transform,
        *typed_mock_args,
        dataspec_type=sp.type_name(dataspec.prototype()),
        dataspec_name=None,
        **typed_named_mock_args,
    )
    mock_constraint(mock)
    attach_variant(dataspec, mock, st.ConstraintKind.MOCK)

    return mock, public_context


def compile_dp(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
    privacy_limit: st.PrivacyLimit,
) -> Tuple[st.DataSpec, Collection[str]]:
    """Simple DP compilation.

    Only check the dataspec's parents, do not go further up in the graph.
    """
    # Current dataspec verifies the constraint?
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            variant_constraint=constraint,
            kind=st.ConstraintKind.DP,
            public_context=public_context,
            privacy_limit=privacy_limit,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                variant_constraint=constraint,
                kind=st.ConstraintKind.DP,
                public_context=public_context,
                privacy_limit=privacy_limit,
            ):
                return variant, public_context

    if not dataspec.is_transformed():
        return compile_synthetic(dataspec_validator, dataspec, public_context)

    transform = dataspec.transform()
    if dataspec.prototype() == sp.Dataset:
        dataset = cast(st.Dataset, dataspec)
        _, DatasetStaticChecker = routing.get_dataset_op(transform)
        is_dp_applicable = DatasetStaticChecker(dataset).is_dp_applicable(
            public_context
        )
        dp_transform = DatasetStaticChecker(dataset).dp_transform()
    else:
        scalar = cast(st.Scalar, dataspec)
        _, ScalarStaticChecker = routing.get_scalar_op(transform)
        is_dp_applicable = ScalarStaticChecker(scalar).is_dp_applicable(
            public_context
        )
        dp_transform = ScalarStaticChecker(scalar).dp_transform()

    if not is_dp_applicable:
        return compile_synthetic(dataspec_validator, dataspec, public_context)

    # Create the DP variant
    assert dp_transform is not None
    budget = privacy_budget(privacy_limit)
    seed = global_context().generate_seed(salt=time_ns())
    args, kwargs = dataspec.parents()
    dp_variant = cast(
        st.DataSpec,
        transformed(
            dp_transform,
            *args,
            dataspec_type=sp.type_name(dataspec.prototype()),
            dataspec_name=None,
            budget=budget,
            seed=seed,
            **kwargs,
        ),
    )
    dp_constraint(
        dataspec=dp_variant,
        required_context=list(public_context),
        privacy_limit=privacy_limit,
    )
    attach_variant(
        original=dataspec,
        variant=dp_variant,
        kind=st.ConstraintKind.DP,
    )

    return dp_variant, public_context



import os
import typing as t

import pyarrow as pa

from sarus_data_spec import typing as st
from sarus_data_spec.constants import CACHE_SCALAR_TASK, TO_PARQUET_TASK
from sarus_data_spec.manager.asyncio.api.api_computation import ApiComputation
from sarus_data_spec.manager.asyncio.base import BaseAsyncManager
from sarus_data_spec.manager.asyncio.worker.arrow_computation import (
    ToArrowComputation,
)
from sarus_data_spec.manager.asyncio.worker.cache_scalar_computation import (
    CacheScalarComputation,
)
from sarus_data_spec.manager.asyncio.worker.caching_computation import (
    ToParquetComputation,
)
from sarus_data_spec.manager.asyncio.worker.schema_computation import (
    SchemaComputation,
)
from sarus_data_spec.manager.asyncio.worker.value_computation import (
    ValueComputation,
)
from sarus_data_spec.manager.ops.asyncio.processor.routing import (
    TransformedDataset,
    TransformedScalar,
)
from sarus_data_spec.manager.ops.asyncio.source.routing import SourceScalar
import sarus_data_spec.manager.typing as smt
import sarus_data_spec.protobuf as sp
import sarus_data_spec.status as stt
import sarus_data_spec.storage.typing as storage_typing

computation_timeout = os.environ.get(
    'DELEGATING_COMPUTATION_TIMEOUT', default=600
)
computation_max_delay = os.environ.get(
    'DELEGATING_COMPUTATION_MAX_DELAY', default=10
)


class DelegatingManager(BaseAsyncManager):
    """Manager that can compute a result locally or delegate a computation.

    This manager is initialized with a reference to the remote Manager so that
    we can register its statuses in the local storage.

    The local computations implementations are taken from the WorkerManager.

    Subclasses should implement the following three computations to compute
    values remotely:
        - self.remote_to_parquet_computation
        - self.remote_to_arrow_computation
        - self.remote_value_computation

    Subclasses also have to implement the `is_remotely_computed` method to
    decide if a computation should be performed locally or remotely.

    Finally, subclasses should implement the `delegate_manager_statuses` method
    to fetch statuses from the remote Manager.
    """

    def __init__(
        self,
        storage: storage_typing.Storage,
        protobuf: sp.Manager,
        remote_manager: smt.Manager,
    ) -> None:
        super().__init__(storage, protobuf)
        self.remote_manager = remote_manager
        self.local_schema_computation = SchemaComputation(self)
        self.local_to_arrow_computation = ToArrowComputation(
            self, ToParquetComputation(self)
        )
        self.local_value_computation = ValueComputation(
            self, CacheScalarComputation(self)
        )
        self.local_to_parquet_computation = ToParquetComputation(self)
        self.local_cache_scalar_computation = CacheScalarComputation(self)

        # To define in subclasses
        self.remote_to_parquet_computation: ApiComputation[None]
        self.remote_to_arrow_computation: ApiComputation[
            t.AsyncIterator[pa.RecordBatch]
        ]
        self.remote_cache_scalar_computation: ApiComputation[t.Any]
        self.remote_value_computation: ApiComputation[t.Any]

    def dataspec_computation(self, dataspec: st.DataSpec) -> smt.Computation:
        """Return the computation for a Dataspec."""
        is_delegated = self.is_delegated(dataspec)
        proto = dataspec.prototype()
        if proto == sp.Dataset:
            if is_delegated:
                return self.remote_to_arrow_computation
            else:
                return self.local_to_arrow_computation
        else:
            if is_delegated:
                return self.remote_value_computation
            else:
                return self.local_value_computation

    def _delegate_manager_status(
        self, dataspec: st.DataSpec, task_name: str
    ) -> t.Optional[st.Status]:
        """Fetch the remote status a single Dataspec."""
        statuses = self._delegate_manager_statuses(
            [dataspec], task_name=task_name
        )
        (status,) = statuses
        return status

    def _delegate_manager_statuses(
        self, dataspecs: t.List[st.DataSpec], task_name: str
    ) -> t.List[t.Optional[st.Status]]:
        """Fetch the remote statuses for a list of Dataspecs."""
        raise NotImplementedError

    def is_delegated(self, dataspec: st.DataSpec) -> bool:
        """Return True is the dataspec's computation is delegated."""
        raise NotImplementedError

    def is_cached(self, dataspec: st.DataSpec) -> bool:
        """Sets whether a dataset should be cached or not"""
        raise NotImplementedError

    def status(
        self, dataspec: st.DataSpec, task_name: str
    ) -> t.Optional[st.Status]:
        """Reads the delegate manager's status and update the API status if
        needed.

        Returns the API manager's status.
        """
        local_status = stt.last_status(
            dataspec=dataspec,
            task=task_name,
            manager=self,
        )

        # TODO think of a finer way to distinguish delegated tasks
        if task_name in [CACHE_SCALAR_TASK, TO_PARQUET_TASK]:
            # cannot delegate the caching
            return local_status

        if not self.is_delegated(dataspec):
            # computation not delegated, return local status
            return local_status

        delegate_status = self._delegate_manager_status(dataspec, task_name)
        if delegate_status is None:
            # return local status which may be pending
            return local_status

        local_stage = local_status.task(task_name) if local_status else None
        delegate_stage = delegate_status.task(task_name)
        assert delegate_stage

        if delegate_stage == local_stage:
            # no progress: update not needed
            return local_status

        # copy delegate manager status as API status
        if delegate_stage.ready():
            status, _ = stt.ready(
                dataspec=dataspec,
                task=task_name,
                manager=self,
                properties=delegate_stage.properties(),
            )
        elif delegate_stage.error():
            status, _ = stt.error(
                dataspec=dataspec,
                task=task_name,
                manager=self,
                properties=delegate_stage.properties(),
            )
        elif delegate_stage.processing():
            status, _ = stt.processing(
                dataspec=dataspec,
                task=task_name,
                manager=self,
                properties=delegate_stage.properties(),
            )
        elif delegate_stage.pending():
            status, _ = stt.pending(
                dataspec=dataspec,
                task=task_name,
                manager=self,
                properties=delegate_stage.properties(),
            )
        else:
            raise ValueError('Delegate manager status not properly managed')

        return status

    async def async_schema(self, dataset: st.Dataset) -> st.Schema:
        """Schema computation is done locally."""
        return await self.local_schema_computation.task_result(
            dataspec=dataset
        )

    async def async_to_parquet(self, dataset: st.Dataset) -> None:
        await self.local_to_parquet_computation.complete_task(dataspec=dataset)

    async def async_cache_scalar(self, scalar: st.Scalar) -> None:
        await self.local_cache_scalar_computation.complete_task(
            dataspec=scalar
        )

    async def async_to_arrow_op(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Route a Dataset to its Op implementation.

        When the computation is not delegated the manager should also be
        able to compute the value.
        """
        if dataset.is_transformed():
            iterator = await TransformedDataset(dataset).to_arrow(
                batch_size=batch_size
            )
            return iterator

        else:
            raise ValueError('Dataset should be transformed.')

    async def async_value_op(self, scalar: st.Scalar) -> t.Any:
        """Route a Scalar to its Op implementation.

        This method is shared between API and Worker because when the data is
        not cached the API manager should also be able to compute the value.
        """
        if scalar.is_transformed():
            return await TransformedScalar(scalar).value()
        else:
            return await SourceScalar(scalar).value()

    async def async_schema_op(self, dataset: st.Dataset) -> st.Schema:
        if dataset.is_transformed():
            return await TransformedDataset(dataset).schema()
        else:
            raise ValueError('Dataset should be transformed.')

    def engine(self, uri: str) -> smt.sa_engine:
        raise NotImplementedError

    def computation_timeout(self, dataspec: st.DataSpec) -> int:
        return int(computation_timeout)

    def computation_max_delay(self, dataspec: st.DataSpec) -> int:
        return int(computation_max_delay)



from __future__ import annotations

from typing import Any, Dict, List, Optional, Type, TypeVar, Union, cast
import base64

from google.protobuf.any_pb2 import Any as AnyProto
from google.protobuf.descriptor_pool import Default
from google.protobuf.json_format import (
    MessageToDict,
    MessageToJson,
    Parse,
    ParseDict,
)
from google.protobuf.message_factory import MessageFactory

from sarus_data_spec.protobuf.proto_container_pb2 import ProtoContainer
from sarus_data_spec.protobuf.typing import Protobuf
import sarus_data_spec as s


def message(type_name: str) -> Protobuf:
    """Return a message instance from a type_name."""
    return MessageFactory().GetPrototype(
        Default().FindMessageTypeByName(type_name)  # type: ignore
    )()


def message_type(type_name: str) -> Type[Protobuf]:
    """Return a message type from a type_name."""
    return MessageFactory().GetPrototype(
        Default().FindMessageTypeByName(type_name)  # type: ignore
    )


def type_name(message: Union[Type[Protobuf], Protobuf]) -> str:
    """Return a type_name from a message."""
    return cast(str, message.DESCRIPTOR.full_name)


def wrap(message: Protobuf) -> AnyProto:
    """Wrap a Message into an Any"""
    wrapped_message: AnyProto = AnyProto()
    wrapped_message.Pack(
        message,
        type_url_prefix=cast(bytes, s.PACKAGE_NAME),
        deterministic=True,
    )
    return wrapped_message


def unwrap(wrapped_message: AnyProto) -> Protobuf:
    """Unwrap an Any to a Message"""
    result = message(wrapped_message.TypeName())  # type: ignore
    wrapped_message.Unpack(result)
    return result


M = TypeVar('M', bound=Protobuf)


def copy(value: M) -> M:
    result = message(value.DESCRIPTOR.full_name)
    result.CopyFrom(value)
    return cast(M, result)


def serialize(message: M) -> bytes:
    return wrap(message).SerializeToString(deterministic=True)


def deserialize(bytes: bytes) -> Protobuf:
    wrapped_message: AnyProto = AnyProto()
    wrapped_message.MergeFromString(bytes)
    return unwrap(wrapped_message)


def json_serialize(message: Protobuf) -> bytes:
    return MessageToJson(
        wrap(message),
        including_default_value_fields=True,
        preserving_proto_field_name=True,
        sort_keys=True,
    ).encode('utf8')


def json_deserialize(bytes: bytes) -> Protobuf:
    wrapped_message: AnyProto = AnyProto()
    Parse(bytes.decode('utf8'), wrapped_message)
    return unwrap(wrapped_message)


def dict_serialize(message: Protobuf) -> Dict[str, Any]:
    return MessageToDict(
        wrap(message),
        including_default_value_fields=True,
        preserving_proto_field_name=True,
    )


def dict_deserialize(dct: Dict[str, Any]) -> Protobuf:
    wrapped_message: AnyProto = AnyProto()
    ParseDict(dct, wrapped_message)
    return unwrap(wrapped_message)


def json(message: Protobuf) -> str:
    return MessageToJson(
        wrap(message),
        including_default_value_fields=True,
        preserving_proto_field_name=True,
        sort_keys=True,
    )


def dejson(string: str) -> Protobuf:
    wrapped_message: AnyProto = AnyProto()
    Parse(string, wrapped_message)
    return unwrap(wrapped_message)


def to_base64(message: Protobuf) -> str:
    return base64.b64encode(serialize(message)).decode('ASCII')


def from_base64(string: str, message: Optional[M] = None) -> M:
    return cast(M, unwrap(AnyProto().FromString(base64.b64decode(string))))


def serialize_protos_list(protos: List[M]) -> str:
    """stores protos in a container and serialize it to a string"""
    return to_base64(
        ProtoContainer(protos=[wrap(element) for element in protos])
    )


def deserialize_proto_list(string: str) -> List[Protobuf]:
    """deserialize ad hoc proto containers to a list of protobufs"""
    proto_cont = from_base64(string, ProtoContainer())
    return [unwrap(element) for element in proto_cont.protos]



from __future__ import annotations

import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.schema import to_arrow
from sarus_data_spec.base import Referring
from sarus_data_spec.constants import DATASET_SLUGNAME, PUBLIC
from sarus_data_spec.path import Path, path
from sarus_data_spec.type import Type
import sarus_data_spec.dataset as sd
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Schema(Referring[sp.Schema]):
    """A python class to describe schemas"""

    def __init__(self, protobuf: sp.Schema, store: bool = True) -> None:
        self._referred = {
            protobuf.dataset
        }  # This has to be defined before it is initialized
        super().__init__(protobuf, store=store)

    def prototype(self) -> t.Type[sp.Schema]:
        """Return the type of the underlying protobuf."""
        return sp.Schema

    def name(self) -> str:
        return self._protobuf.name

    def dataset(self) -> sd.Dataset:
        return t.cast(
            sd.Dataset, self.storage().referrable(self._protobuf.dataset)
        )

    def to_arrow(self) -> pa.Schema:
        return to_arrow(self.protobuf())

    def type(self) -> Type:
        """Returns the first type level of the schema"""
        return Type(self.protobuf().type)

    def data_type(self) -> Type:
        """Returns the first type level containing the data,
        hence skips the protected_entity struct if there is one"""
        return self.type().data_type()

    def is_protected(self) -> bool:
        return self.type().has_protected_format()

    def protected_path(self) -> Path:
        """Returns the path to the protected entities"""
        return Path(self.protobuf().protected)

    # TODO: Add to_parquet, to_tensorflow, to_sql... here?
    # The Schema has a manager, it would provide the implementation

    def tables(self) -> t.List[st.Path]:
        struct_paths = self.data_type().structs()
        if struct_paths is None:  # there is no struct
            return []
        if len(struct_paths) == 0:  # struct is the first level
            return [path(label='data')]
        return [
            path(label='data', paths=[t.cast(Path, element)])
            for element in struct_paths
        ]

    def private_tables(self) -> t.List[st.Path]:
        return [
            table
            for table in self.tables()
            if self.data_type().sub_types(table)[0].properties()[PUBLIC]
            != str(True)
        ]

    def public_tables(self) -> t.List[st.Path]:
        return [
            table
            for table in self.tables()
            if self.data_type().sub_types(table)[0].properties()[PUBLIC]
            == str(True)
        ]


# Builder
def schema(
    dataset: st.Dataset,
    fields: t.Optional[t.Mapping[str, st.Type]] = None,
    schema_type: t.Optional[st.Type] = None,
    protected_paths: t.Optional[sp.Path] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
    name: t.Optional[str] = None,
) -> Schema:
    """A builder to ease the construction of a schema"""
    if name is None:
        name = dataset.properties().get(
            DATASET_SLUGNAME, f'{dataset.name()}_schema'
        )
    assert name is not None

    if fields is not None:
        return Schema(
            sp.Schema(
                dataset=dataset.uuid(),
                name=name,
                type=sp.Type(
                    struct=sp.Type.Struct(
                        fields=[
                            sp.Type.Struct.Field(
                                name=name, type=type.protobuf()
                            )
                            for name, type in fields.items()
                        ]
                    )
                ),
                protected=protected_paths,
                properties=properties,
            )
        )
    if schema_type is not None:
        return Schema(
            sp.Schema(
                dataset=dataset.uuid(),
                name=name,
                type=schema_type.protobuf(),
                protected=protected_paths,
                properties=properties,
            )
        )
    # If none of fields or type is defined, set type to Null
    return Schema(
        sp.Schema(
            dataset=dataset.uuid(),
            name=name,
            type=sp.Type(name='Null', null=sp.Type.Null()),
            protected=protected_paths,
            properties=properties,
        )
    )


if t.TYPE_CHECKING:
    test_schema: st.Schema = schema(sd.sql(uri='sqlite:///:memory:'))



from typing import AsyncIterator, List, Union
import typing as t

import numpy as np
import pyarrow as pa

from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.asyncio.utils import async_iter
from sarus_data_spec.manager.ops.asyncio.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.marginals import marginals as marginals_builder
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st

try:
    from sarus_data_spec.manager.ops.asyncio.processor.standard.sampling.sql import (  # noqa: E501
        sql_sample_to_arrow,
    )
except ModuleNotFoundError as exception:
    # for the public repo
    if (
        exception.name
        == 'sarus_data_spec.manager.ops.asyncio.processor.standard.sampling'  # noqa: E501
    ):
        pass
    else:
        raise exception


try:
    from sarus_statistics.tasks.size.sample_visitor import sampled_size
except ModuleNotFoundError as exception:
    # for the public repo
    if 'sarus_statistics' in str(exception.name):
        pass
    else:
        raise exception


class SampleStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            protected_paths=parent_schema.protobuf().protected,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Sample(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a sample transform
    """

    async def size(self) -> st.Size:
        parent_size = await self.parent_size()
        size_ratio = new_sampling_ratio(self.dataset, parent_size.statistics())
        if size_ratio >= 1:
            return parent_size
        return size_builder(
            self.dataset, sampled_size(parent_size.statistics(), size_ratio)
        )

    async def bounds(self) -> st.Bounds:
        parent_bounds = await self.parent_bounds()
        size_ratio = new_sampling_ratio(
            self.dataset, parent_bounds.statistics()
        )
        if size_ratio >= 1:
            return parent_bounds
        return bounds_builder(
            self.dataset, sampled_size(parent_bounds.statistics(), size_ratio)
        )

    async def marginals(self) -> st.Marginals:
        parent_marginals = await self.parent_marginals()
        size_ratio = new_sampling_ratio(
            self.dataset, parent_marginals.statistics()
        )
        if size_ratio >= 1:
            return parent_marginals
        return marginals_builder(
            self.dataset,
            sampled_size(parent_marginals.statistics(), size_ratio),
        )

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = t.cast(Dataset, self.parent())
        if parent.manager().is_big_data(parent):
            return await sql_sample_to_arrow(self.dataset, batch_size)
        return await self._arrow_to_arrow(batch_size)

    async def _arrow_to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await sample_arrow_to_arrow(self, batch_size)


# Copied from
# https://github.com/huggingface/datasets/blob/master/src/datasets/table.py
def fast_gather(
    indices: Union[List[int], np.ndarray],
    batches: List[pa.RecordBatch],
    batch_size: int,
) -> AsyncIterator[pa.RecordBatch]:
    """
    Create a pa.Table by gathering the records at the records at the specified
    indices. Should be faster than
    pa.concat_tables(
        table.fast_slice(int(i) % table.num_rows, 1) for i in indices
    )
    since NumPy can compute the binary searches in parallel,
    highly optimized C
    """
    assert len(indices), "Indices must be non-empty"
    offsets: np.ndarray = np.cumsum(
        [0] + [len(b) for b in batches], dtype=np.int64
    )
    batch_indices = np.searchsorted(offsets, indices, side="right") - 1
    # it is important to combine the chunks of the table: if each record batch
    # slice has a length smaller than batch_size, then they will be
    # kept as they are and we will not get batches of size `batch_size`
    return async_iter(
        pa.Table.from_batches(
            [
                batches[batch_idx].slice(i - offsets[batch_idx], 1)
                for batch_idx, i in zip(batch_indices, indices)
            ],
            schema=batches[0].schema,
        )
        .combine_chunks()
        .to_batches(max_chunksize=batch_size)
    )


def new_sampling_ratio(
    dataset: st.Dataset, statistics: st.Statistics
) -> float:
    """From a transformed dataset which last transform is sample or
    differentiated_sample, get sampling ratio"""
    spec = dataset.transform().protobuf().spec
    transform_type = spec.WhichOneof('spec')
    assert transform_type
    sampling_spec = getattr(spec, transform_type)
    if sampling_spec.HasField('fraction'):
        return t.cast(float, sampling_spec.fraction)
    return t.cast(float, sampling_spec.size / statistics.size())


async def sample_arrow_to_arrow(
    op: StandardDatasetImplementation, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:
    spec = op.dataset.transform().protobuf().spec
    transform_type = spec.WhichOneof('spec')
    assert transform_type
    sampling_spec = getattr(spec, transform_type)
    seed = Scalar(sampling_spec.seed).value()
    generator = np.random.default_rng(seed)
    parent_batches = [batch async for batch in await op.parent_to_arrow()]
    parent_table = pa.Table.from_batches(parent_batches)

    if sampling_spec.HasField('fraction'):
        new_size = int(sampling_spec.fraction * parent_table.num_rows)
    else:
        new_size = sampling_spec.size

    parent_size = await op.parent_size()
    if new_size >= parent_size.statistics().size():
        return await op.parent_to_arrow(batch_size)

    indices = generator.choice(
        parent_table.num_rows,
        replace=False,
        size=new_size,
    )
    return fast_gather(
        indices=indices,
        batches=parent_table.to_batches(max_chunksize=1000),
        batch_size=batch_size,
    )



import asyncio
import logging
import typing as t

import pyarrow as pa

from sarus_data_spec.manager.asyncio.base import BaseComputation, T
from sarus_data_spec.manager.typing import Manager
import sarus_data_spec.status as stt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)

# This is done to avoid circular dependencies..


class TypedWorkerManager(Manager, t.Protocol):
    async def async_value_op(self, scalar: st.Scalar) -> st.DataSpecValue:
        ...

    async def async_schema_op(self, dataset: st.Dataset) -> st.Schema:
        ...

    async def async_to_arrow_op(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    def computation_timeout(self, dataspec: st.DataSpec) -> int:
        ...

    def computation_max_delay(self, dataspec: st.DataSpec) -> int:
        ...


class WorkerComputation(BaseComputation[T]):
    def __init__(self, manager: TypedWorkerManager):
        self._manager: TypedWorkerManager = manager

    def manager(self) -> TypedWorkerManager:
        return self._manager

    async def read_ready_result(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> T:
        raise NotImplementedError

    async def processing(self, dataspec: st.DataSpec) -> st.Status:
        """If processing, wait for the task to be ready.
        Such a case can happen if another manager has taken the computation
        of the task. After a given timeout, an error is raised.
        """

        stage = await self.wait_for_computation(
            dataspec=dataspec,
            current_stage='processing',
            timeout=self.manager().computation_timeout(dataspec),
            max_delay=self.manager().computation_max_delay(dataspec),
        )
        if stage.processing():
            stt.error(
                dataspec=dataspec,
                manager=dataspec.manager(),
                task=self.task_name,
                properties={
                    "message": "TimeOutError:Processing time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                    'relaunch': str(True),
                },
            )
            raise stt.DataSpecErrorStatus(
                (
                    True,
                    "TimeOutError:Processing time out for task"
                    f" {self.task_name} on dataspec {dataspec}",
                )
            )
        # if the stage is an error, it is complete_task
        # that decides what to do via the error_policy
        return await self.complete_task(dataspec)

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        status = stt.last_status(
            dataspec=dataspec, manager=self.manager(), task=self.task_name
        )
        if status is None:
            _, is_updated = stt.processing(
                dataspec=dataspec,
                manager=self.manager(),
                task=self.task_name,
            )
            if is_updated:
                return asyncio.create_task(
                    self.prepare(dataspec),
                    name=self.task_name + dataspec.uuid(),
                )

        return None

    async def prepare(self, dataspec: st.DataSpec) -> None:
        """Prepare the computation and set the status accordingly.

        It is up to the computation to define what preparing means. It can be
        computing and caching the data as well as simply checking that the
        parents are themselves ready.
        """
        raise NotImplementedError



from enum import Enum

import pandas as pd

from sarus_data_spec import typing as st

DATA = 'data'
USER_COLUMN = 'sarus_protected_entity'
WEIGHTS = 'sarus_weights'
PUBLIC = 'sarus_is_public'

# constants used in protection stored in each struct
NON_EMPTY_PROTECTED_PATHS = 'non_zero_protected_values'
STRUCT_KIND = 'merge_paths'
TO_MERGE = 'fks_for_merging'

# protection constants for paths in types
LIST_VALUES = 'sarus_list_values'
ARRAY_VALUES = 'sarus_array_values'
OPTIONAL_VALUE = 'sarus_optional_value'
CONSTRAINED_VALUE = 'sarus_constrained_value'


class StructKind(Enum):
    HAS_PE = '0'
    NO_PE = '1'
    TO_MERGE = '2'


# constants for type properties
TEXT_MIN_LENGTH = 'min_length'
TEXT_MAX_LENGTH = 'max_length'
TEXT_CHARSET = 'text_char_set'
TEXT_EXACT_CHARSET = 'FullUserInput'
TEXT_ALPHABET_NAME = 'text_alphabet_name'
SQL_MAPPING = 'sql_mapping'
FLOAT_DISTRIBUTION = 'distribution_model'
FLOAT_DIST_PARAMS = 'parameters'
MAX_MAX_MULT = 'max_max_multiplicity'

# constants for dataset properties
DATASET_SLUGNAME = 'slugname'

# constants for schema properties
PRIMARY_KEYS = 'primary_keys'
FOREIGN_KEYS = 'foreign_keys'

# names when transforming datetime type in struct
DATETIME_YEAR = 'year'
DATETIME_MONTH = 'month'
DATETIME_DAY = 'day'
DATETIME_HOUR = 'hour'
DATETIME_MINUTES = 'minutes'
DATETIME_SECONDS = 'seconds'


# sql, to_sql status
TO_SQL_TASK = "sql_preparation"
SQL_TASK = 'sql'

# Big Data Status
BIG_DATA_TASK = 'big_data_dataset'
BIG_DATA_THRESHOLD = 'threshold'
IS_BIG_DATA = 'is_big_data'
DATASET_N_LINES = 'dataset_n_lines'
DATASET_N_BYTES = 'dataset_n_bytes'
THRESHOLD_TYPE = "threshold_type"

# Caching Status
TO_PARQUET_TASK = 'to_parquet'
CACHE = TO_PARQUET_TASK
CACHE_PATH = 'path'
COMPUTATION_QUEUED = 'computation_queued'
TO_SQL_CACHING_TASK = 'to_sql_caching'
SQL_CACHING_URI = 'sql_caching_uri'
TABLE_MAPPING = 'table_mapping'
EXTENDED_TABLE_MAPPING = 'extended_table_mapping'

# Caching infos for scalar
CACHE_TYPE = 'cache_type'
CACHE_PROTO = 'cache_proto'
SCALAR_TASK = 'scalar_value'


class ScalarCaching(Enum):
    PICKLE = 'pickle'
    PROTO = 'protobuf'


# Attributes Status
SCHEMA_TASK = 'schema'
SIZE_TASK = 'size'
BOUNDS_TASK = 'bounds'
MARGINALS_TASK = 'marginals'
ARROW_TASK = 'arrow'
PROTECTION_TASK = 'protection_task'
USER_SETTINGS_TASK = 'user_settings_task'
PUBLIC_TASK = 'public_task'
CACHE_SCALAR_TASK = 'cache_scalar'
QB_TASK = 'query_builder'
LINKS_TASK = 'links_statistics'
SYNTHETIC_TASK = 'synthetic'


# Privacy
PEP_TOKEN = "pep_token"
PRIVACY_LIMIT = "privacy_limit"
CONSTRAINT_KIND = "constraint_kind"
BEST_ALTERNATIVE = "best_alternative"

# QUERYBUILDER
QUERIES = 'queries'

# Attributes names
PRIVATE_QUERY = "private_query"
IS_REMOTE = "is_remote"
VARIANT_UUID = "variant_uuid"
RELATIONSHIP_SPEC = "relationship_spec"


# SYNTHETIC DATA
SYNTHETIC_MODEL = 'sarus_synthetic_model'
TRAIN_CORRELATIONS = 'train_correlations'


class SyntheticDataSettings:
    """Namespace for SD generation settings"""

    BATCH_SIZE = 64
    EPOCHS = 10


# map between sqlalchemy dialect names to sarus notion of dialects
SQLALCHEMY_DIALECT_MAP = {
    "postgresql": st.SQLDialect.POSTGRES,
    "mssql": st.SQLDialect.SQL_SERVER,
    "mysql": st.SQLDialect.MY_SQL,
    "sqlite": st.SQLDialect.SQLLITE,
    "oracle": st.SQLDialect.ORACLE,
    "bigquery": st.SQLDialect.BIG_QUERY,
    "redshift": st.SQLDialect.REDSHIFT,
    "hive": st.SQLDialect.HIVE,
}

# List of types a Dataset can be converted to
DATASET_TYPES = [pd.DataFrame]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/size.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import statistics_pb2 as sarus__data__spec_dot_protobuf_dot_statistics__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/size.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#sarus_data_spec/protobuf/size.proto\x12\x0fsarus_data_spec\x1a)sarus_data_spec/protobuf/statistics.proto\"\xd2\x01\n\x04Size\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12/\n\nstatistics\x18\x04 \x01(\x0b\x32\x1b.sarus_data_spec.Statistics\x12\x39\n\nproperties\x18\x05 \x03(\x0b\x32%.sarus_data_spec.Size.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[sarus__data__spec_dot_protobuf_dot_statistics__pb2.DESCRIPTOR,])




_SIZE_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Size.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Size.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Size.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=261,
  serialized_end=310,
)

_SIZE = _descriptor.Descriptor(
  name='Size',
  full_name='sarus_data_spec.Size',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='sarus_data_spec.Size.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='sarus_data_spec.Size.dataset', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Size.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='statistics', full_name='sarus_data_spec.Size.statistics', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Size.properties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SIZE_PROPERTIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=310,
)

_SIZE_PROPERTIESENTRY.containing_type = _SIZE
_SIZE.fields_by_name['statistics'].message_type = sarus__data__spec_dot_protobuf_dot_statistics__pb2._STATISTICS
_SIZE.fields_by_name['properties'].message_type = _SIZE_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['Size'] = _SIZE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Size = _reflection.GeneratedProtocolMessageType('Size', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _SIZE_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.size_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Size.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _SIZE,
  '__module__' : 'sarus_data_spec.protobuf.size_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Size)
  })
_sym_db.RegisterMessage(Size)
_sym_db.RegisterMessage(Size.PropertiesEntry)


_SIZE_PROPERTIESENTRY._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Any, Dict
import logging
import pickle as pkl
import typing as t

from sarus_data_spec.manager.ops.asyncio.base import ScalarImplementation
import sarus_data_spec.protobuf as sp

logger = logging.getLogger(__name__)

try:
    import xgboost  # type: ignore[import]
except ModuleNotFoundError:
    logger.info("XGBoost not installed, XGBoost models not available.")

try:
    from sklearn import (
        cluster,
        decomposition,
        ensemble,
        model_selection,
        preprocessing,
        svm,
    )
except ModuleNotFoundError:
    logger.info("sklearn not installed, sklearn models not available.")


class Model(ScalarImplementation):
    async def value(self) -> t.Any:
        model_spec = self.scalar.protobuf().spec.model
        model_code_name = sp.Scalar.Model.ModelClass.Name(
            model_spec.model_class
        )

        args = pkl.loads(model_spec.arguments)
        kwargs = pkl.loads(model_spec.named_arguments)

        if model_code_name.startswith("SK_"):
            model_mapping: Dict[str, Any] = {
                "SK_SVC": svm.SVC,
                "SK_ONEHOT": preprocessing.OneHotEncoder,
                "SK_LABEL_ENCODER": preprocessing.LabelEncoder,
                "SK_PCA": decomposition.PCA,
                # cluster
                "SK_AFFINITY_PROPAGATION": cluster.AffinityPropagation,
                "SK_AGGLOMERATIVE_CLUSTERING": cluster.AgglomerativeClustering,
                "SK_BIRCH": cluster.Birch,
                "SK_DBSCAN": cluster.DBSCAN,
                "SK_FEATURE_AGGLOMERATION": cluster.FeatureAgglomeration,
                "SK_KMEANS": cluster.KMeans,
                "SK_MINIBATCH_KMEANS": cluster.MiniBatchKMeans,
                "SK_MEAN_SHIFT": cluster.MeanShift,
                "SK_OPTICS": cluster.OPTICS,
                "SK_SPECTRAL_CLUSTERING": cluster.SpectralClustering,
                "SK_SPECTRAL_BICLUSTERING": cluster.SpectralBiclustering,
                "SK_SPECTRAL_COCLUSTERING": cluster.SpectralCoclustering,
                # ensemble
                "SK_ADABOOST_CLASSIFIER": ensemble.AdaBoostClassifier,
                "SK_ADABOOST_REGRESSOR": ensemble.AdaBoostRegressor,
                "SK_BAGGING_CLASSIFIER": ensemble.BaggingClassifier,
                "SK_BAGGING_REGRESSOR": ensemble.BaggingRegressor,
                "SK_EXTRA_TREES_CLASSIFIER": ensemble.ExtraTreesClassifier,
                "SK_EXTRA_TREES_REGRESSOR": ensemble.ExtraTreesRegressor,
                "SK_GRADIENT_BOOSTING_CLASSIFIER": ensemble.GradientBoostingClassifier,
                "SK_GRADIENT_BOOSTING_REGRESSOR": ensemble.GradientBoostingRegressor,
                "SK_ISOLATION_FOREST": ensemble.IsolationForest,
                "SK_RANDOM_FOREST_CLASSIFIER": ensemble.RandomForestClassifier,
                "SK_RANDOM_FOREST_REGRESSOR": ensemble.RandomForestRegressor,
                "SK_RANDOM_TREES_EMBEDDING": ensemble.RandomTreesEmbedding,
                "SK_STACKING_CLASSIFIER": ensemble.StackingClassifier,
                "SK_STACKING_REGRESSOR": ensemble.StackingRegressor,
                "SK_VOTING_CLASSIFIER": ensemble.VotingClassifier,
                "SK_VOTING_REGRESSOR": ensemble.VotingRegressor,
                "SK_HIST_GRADIENT_BOOSTING_CLASSIFIER": ensemble.HistGradientBoostingClassifier,
                "SK_HIST_GRADIENT_BOOSTING_REGRESSOR": ensemble.HistGradientBoostingRegressor,
                # model_selection
                "SK_REPEATED_STRATIFIED_KFOLD": model_selection.RepeatedStratifiedKFold,
                "SK_KFOLD": model_selection.KFold,
            }
        elif model_code_name.startswith("XGB_"):
            model_mapping = {
                "XGB_CLASSIFIER": xgboost.XGBClassifier,
            }
        else:
            raise ValueError(
                f"Model {model_code_name} is not supported by sarus."
            )

        ModelClass = model_mapping[model_code_name]
        return ModelClass(*args, **kwargs)



from typing import Collection, List, Optional, Tuple, cast
import logging

import sarus_data_spec.typing as st

ArgStruct = Tuple[List[int], List[str]]
logger = logging.getLogger(__name__)


def verifies(
    variant_constraint: st.VariantConstraint,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    privacy_limit: Optional[st.PrivacyLimit],
) -> bool:
    if kind == st.ConstraintKind.PUBLIC:
        return verifies_public(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.SYNTHETIC:
        return verifies_synthetic(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.MOCK:
        return verifies_mock(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.DP:
        return verifies_dp(
            variant_constraint=variant_constraint,
            privacy_limit=privacy_limit,
        )

    else:  # kind == st.ConstraintKind.PEP:
        return verifies_pep(variant_constraint=variant_constraint)


def verifies_public(variant_constraint: st.VariantConstraint) -> bool:
    return variant_constraint.constraint_kind() == st.ConstraintKind.PUBLIC


def verifies_synthetic(variant_constraint: st.VariantConstraint) -> bool:
    return variant_constraint.constraint_kind() in [
        st.ConstraintKind.PUBLIC,
        st.ConstraintKind.SYNTHETIC,
    ]


def verifies_mock(variant_constraint: st.VariantConstraint) -> bool:
    return variant_constraint.constraint_kind() in [
        st.ConstraintKind.PUBLIC,
        st.ConstraintKind.MOCK,
    ]


def verifies_pep(
    variant_constraint: st.VariantConstraint,
) -> bool:
    """If we attached a PEP constraint to a dataspec then it is PEP.

    NB: for now we don't check the context nor the privacy limit
    """
    return variant_constraint.constraint_kind() == st.ConstraintKind.PEP


def verifies_dp(
    variant_constraint: st.VariantConstraint,
    privacy_limit: Optional[st.PrivacyLimit],
) -> bool:
    """Check if a variant constraint satisfies a DP profile.

    For now, return True only for strict equality.
    """
    if privacy_limit is None:
        raise ValueError(
            "Input privacy limit required when checking against DP."
        )

    kind = variant_constraint.constraint_kind()
    if kind != st.ConstraintKind.DP:
        return False

    constraint_privacy_limit = variant_constraint.privacy_limit()
    if constraint_privacy_limit is None:
        raise ValueError(
            "Found a DP constraint without a privacy limit "
            "when checking against DP."
        )

    return cast(
        bool,
        privacy_limit.delta_epsilon_dict()
        == constraint_privacy_limit.delta_epsilon_dict(),
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/proto_container.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/proto_container.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.sarus_data_spec/protobuf/proto_container.proto\x12\x0fsarus_data_spec\x1a\x19google/protobuf/any.proto\"\xae\x01\n\x0eProtoContainer\x12$\n\x06protos\x18\x01 \x03(\x0b\x32\x14.google.protobuf.Any\x12\x43\n\nproperties\x18\x02 \x03(\x0b\x32/.sarus_data_spec.ProtoContainer.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_PROTOCONTAINER_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.ProtoContainer.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.ProtoContainer.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.ProtoContainer.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=220,
  serialized_end=269,
)

_PROTOCONTAINER = _descriptor.Descriptor(
  name='ProtoContainer',
  full_name='sarus_data_spec.ProtoContainer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='protos', full_name='sarus_data_spec.ProtoContainer.protos', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.ProtoContainer.properties', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROTOCONTAINER_PROPERTIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=95,
  serialized_end=269,
)

_PROTOCONTAINER_PROPERTIESENTRY.containing_type = _PROTOCONTAINER
_PROTOCONTAINER.fields_by_name['protos'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_PROTOCONTAINER.fields_by_name['properties'].message_type = _PROTOCONTAINER_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['ProtoContainer'] = _PROTOCONTAINER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProtoContainer = _reflection.GeneratedProtocolMessageType('ProtoContainer', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROTOCONTAINER_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.proto_container_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.ProtoContainer.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _PROTOCONTAINER,
  '__module__' : 'sarus_data_spec.protobuf.proto_container_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.ProtoContainer)
  })
_sym_db.RegisterMessage(ProtoContainer)
_sym_db.RegisterMessage(ProtoContainer.PropertiesEntry)


_PROTOCONTAINER_PROPERTIESENTRY._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.constants import DATA, PUBLIC, USER_COLUMN, WEIGHTS
from sarus_data_spec.dataspec_validator.typing import PEPKind
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class ExternalOpImplementation:
    """External PEP op implementation class.

    This class wraps together several elements of an external op
    implementation:
        - `call` is the function that computes the output value from the
          input(s) value(s).
    """

    transform_id: str
    _dp_equivalent: t.Optional[t.Type[ExternalOpImplementation]] = None
    allowed_pep_args: t.List[t.Set[str]] = []

    def dp_equivalent(self) -> t.Optional[ExternalOpImplementation]:
        if not self._dp_equivalent:
            return None
        return self._dp_equivalent()

    @staticmethod
    async def call(*args: t.Any, **kwargs: t.Any) -> t.Any:
        raise NotImplementedError

    def pep_kind(self, *args: t.Any, **kwargs: t.Any) -> PEPKind:
        """Return the PEP properties of the transform.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, it is not PEP if we are aggregating
        the rows (axis=0) and it is PEP if we are aggregating the columns
        (axis=1).

        NB: This function should have the same signature as the call function.
        """
        # Default implementation
        return PEPKind.NOT_PEP

    def is_dp(self, *args: t.Any, **kwargs: t.Any) -> bool:
        """Return True if the DP transform is compatible with the arguments.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, if we are aggregating the rows
        (axis=0), then there might be an equivalent DP transform but if we are
        aggregating the columns there might not (axis=1).

        NB: This function should have the same signature as the call function.
        """
        return False

    async def private_queries(
        self,
        *args: t.Any,
        budget: t.Optional[st.Scalar] = None,
        seed: t.Optional[st.Scalar] = None,
        **kwargs: t.Any,
    ) -> t.List[st.PrivateQuery]:
        """Takes as input the args of the transform (static and dynamic)."""
        if budget is None or seed is None:
            return []
        # Evaluate budget and seed
        budget_value = await budget.async_value()
        seed_value = await seed.async_value()
        queries, _ = await self.private_queries_and_task(
            *args, budget=budget_value, seed=seed_value, **kwargs
        )
        return queries

    @staticmethod
    async def private_queries_and_task(
        *args: t.Any, **kwargs: t.Any
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        raise NotImplementedError


async def arguments_values(
    *args: t.Any, **kwargs: t.Any
) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
    """Evaluate sarus dataspecs among arguments.

    Compute the value of Dataspec arguments.
    """
    data_pe_args = [
        await eval_dataspec_and_exract_protected_entity(arg) for arg in args
    ]
    data_args = [data for data, _ in data_pe_args]

    data_pe_kwargs = {
        name: await eval_dataspec_and_exract_protected_entity(arg)
        for name, arg in kwargs.items()
    }
    data_kwargs = {name: data for name, (data, _) in data_pe_kwargs.items()}

    return data_args, data_kwargs


async def arguments_values_and_protected_entity(
    *args: t.Any, **kwargs: t.Any
) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any], pd.DataFrame]:
    """Evaluate sarus dataspecs and extract the PE.

    Compute the value of Dataspec arguments. Extract all the protections,
    check that they are equal and return the PE value.
    """
    data_pe_args = [
        await eval_dataspec_and_exract_protected_entity(arg) for arg in args
    ]
    data_args = [data for data, _ in data_pe_args]
    pe_args = [pe for _, pe in data_pe_args]

    data_pe_kwargs = {
        name: await eval_dataspec_and_exract_protected_entity(arg)
        for name, arg in kwargs.items()
    }
    data_kwargs = {name: data for name, (data, _) in data_pe_kwargs.items()}
    pe_kwargs = [pe for _, pe in data_pe_kwargs.values()]

    input_protected_entities = list(
        filter(lambda x: x is not None, pe_args + pe_kwargs)
    )
    protected_entity = validate_protected_entity(input_protected_entities)

    return data_args, data_kwargs, protected_entity


def validate_protected_entity(
    protected_entities: t.List[pd.DataFrame],
) -> pd.DataFrame:
    """Check that all protected entities are equal."""
    # If we reach this part then there should be only one input PE
    pe = next(iter(protected_entities), None)
    if pe is None:
        raise ValueError("The dataset was infered PEP but has no input PE")
    # For now, PE in external ops are only viewed as pd.DataFrames
    if not all([candidate.equals(pe) for candidate in protected_entities]):
        raise ValueError(
            "The dataset is PEP but has several differing input PE values"
        )
    return pe


def output_protected_entity(
    input_pe: pd.DataFrame, output_value: pd.DataFrame
) -> pd.DataFrame:
    """Compute the output protected entity of an external transform."""
    # We guarantee that the data.index is a reliable way to trace how
    # the rows were rearranged
    return input_pe.loc[output_value.index]


async def eval_dataspec_and_exract_protected_entity(
    x: t.Union[t.Any, st.DataSpec]
) -> t.Tuple[t.Any, t.Optional[pd.DataFrame]]:
    """Compute the value of a DataSpec and extract its PE.

    Return None for the PE if it is not defined for the argument.
    """
    if not isinstance(x, st.DataSpec):
        return x, None
    if x.prototype() == sp.Dataset:
        dataset = t.cast(st.Dataset, x)
        df_with_pe = await dataset.async_to_pandas()
        return dataframe_split_data_and_protected_entity(df_with_pe)
    else:
        scalar = t.cast(st.Scalar, x)
        return await scalar.async_value(), None


def dataframe_split_data_and_protected_entity(
    raw_data: pd.DataFrame,
) -> t.Tuple[pd.DataFrame, t.Optional[pd.DataFrame]]:
    """Extract the protected entity from a pd.DataFrame.

    Return None for the PE is no PE is found on the DataFrame.
    """
    # TODO use `dataset.is_protected()` but we need the schema
    # for that, and it is not yet available in the SDK
    is_protected = set(raw_data.columns) == {
        PUBLIC,
        USER_COLUMN,
        WEIGHTS,
        DATA,
    }
    if is_protected:
        data = pd.DataFrame.from_records(
            raw_data[DATA].values, index=raw_data.index
        )
        pe = raw_data[[PUBLIC, USER_COLUMN, WEIGHTS]]
    else:
        data = raw_data
        pe = None
    return data, pe


def to_pyarrow_table_with_protected_entity(
    data: pd.DataFrame, pe: t.Optional[pd.DataFrame]
) -> pa.Table:
    """Merge a protection and its data in pandas.

    NB: we return directly the pa.Table otherwise the data column order is lost
    when doing pa.Table.from_pandas(df).to_pandas(). Another solution to
    preserve the column order would be to also return the Table schema and pass
    it as argument to pa.Table.from_pandas(df, schema)
    """
    if pe is not None:
        # Work with Arrow to preserve column order (vs Python dict)
        pe_table = pa.Table.from_pandas(pe)
        data_table = pa.Table.from_pandas(data)
        data_arrays = [
            chunked_array.combine_chunks()
            for chunked_array in data_table.columns
        ]
        data_array = pa.StructArray.from_arrays(
            data_arrays, names=data_table.column_names
        )
        result = pe_table.append_column(DATA, data_array)
    else:
        # TODO also wrap the data in an empty protection
        result = pa.Table.from_pandas(data)
    return result



from __future__ import annotations

from typing import Any

try:
    import skopt
except ModuleNotFoundError:
    pass  # error message in typing.py


async def skopt_bayes_search_cv(*args: Any, **kwargs: Any) -> Any:
    return skopt.BayesSearchCV(*args, **kwargs)



from __future__ import annotations

import asyncio
import logging
import os
import traceback
import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.manager.asyncio.utils import sync, sync_iterator
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.ops.asyncio.processor.external.external_op import (  # noqa: E501
    async_compute_external_value,
)
from sarus_data_spec.manager.typing import Computation, Manager
from sarus_data_spec.schema import Schema
import sarus_data_spec.protobuf as sp
import sarus_data_spec.status as stt
import sarus_data_spec.typing as st

try:
    from sarus_data_spec.bounds import Bounds
    from sarus_data_spec.links import Links
    from sarus_data_spec.marginals import Marginals
    from sarus_data_spec.size import Size

except ModuleNotFoundError:
    pass
try:
    import tensorflow as tf

    from sarus_data_spec.manager.ops.asyncio.tensorflow.features import (
        deserialize,
        flatten,
        nest,
        serialize,
        to_internal_signature,
    )
    from sarus_data_spec.manager.ops.asyncio.tensorflow.tensorflow_visitor import (  # noqa: E501
        convert_tensorflow,
    )
except ModuleNotFoundError:
    pass  # error message printed from typing.py


logger = logging.getLogger(__name__)

BATCH_SIZE = 10000


class BaseAsyncManager(Base):
    """Asynchronous Manager Base implementation
    Make synchronous methods rely on asynchronous ones for consistency.
    """

    def dataspec_computation(self, dataspec: st.DataSpec) -> Computation:
        """Return the Computation for getting the dataspec's value. If sql is
        is true, the sqlcomputation will be returned.
        """
        raise NotImplementedError

    def sql_computation(self) -> Computation:
        """Returns the SQL Computation of the manager. The aim of the method is
        is to reduce code duplication between API and Worker manager
        """
        raise NotImplementedError

    def prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready."""
        sync(self.async_prepare(dataspec))

    async def async_prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready asynchronously."""
        computation = self.dataspec_computation(dataspec)
        await computation.complete_task(dataspec)

    async def async_prepare_parents(self, dataspec: st.DataSpec) -> None:
        """Prepare all the parents of a Dataspec."""
        args, kwargs = dataspec.parents()
        parents = list(args) + list(kwargs.values())
        coros = [self.async_prepare(parent) for parent in parents]
        # here, if many parents potentially fail, we want to be sure
        # that all of them do it, and not only the first one (to
        # modify all statuses accordingly).
        # After, we only raise the first exception for the child.

        results = await asyncio.gather(*coros, return_exceptions=True)
        exceptions = [
            element for element in results if isinstance(element, Exception)
        ]
        if len(exceptions) == 0:
            return
        raise error_aggregation(exceptions)

    async def async_value(self, scalar: st.Scalar) -> t.Any:
        """Reads asynchronously value of a scalar."""
        computation = self.dataspec_computation(scalar)
        return await computation.task_result(dataspec=scalar)

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        return sync(self.async_value(scalar=scalar))

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Reads asynchronous iterator of datast batches"""
        computation = t.cast(
            BaseComputation[t.AsyncIterator[pa.RecordBatch]],
            self.dataspec_computation(dataset),
        )
        return await computation.task_result(
            dataspec=dataset, batch_size=batch_size
        )

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.Iterator[pa.RecordBatch]:
        return sync_iterator(
            self.async_to_arrow(dataset=dataset, batch_size=batch_size)
        )

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        batches_async_it = await self.async_to_arrow(
            dataset=dataset, batch_size=BATCH_SIZE
        )
        arrow_batches = [batch async for batch in batches_async_it]
        return pa.Table.from_batches(arrow_batches).to_pandas()

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        return sync(self.async_to_pandas(dataset=dataset))

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        return sync(self.async_to_tensorflow(dataset=dataset))

    def schema(self, dataset: st.Dataset) -> Schema:
        return t.cast(Schema, sync(self.async_schema(dataset=dataset)))

    def to_parquet(self, dataset: st.Dataset) -> None:
        sync(self.async_to_parquet(dataset=dataset))

    def to_sql(self, dataset: st.Dataset) -> None:
        sync(self.async_to_sql(dataset=dataset))

    def sql_prepare(self, dataset: st.Dataset) -> None:
        """SQL prepare dataset."""
        sync(self.async_sql_prepare(dataset))

    async def async_sql_prepare(self, dataset: st.Dataset) -> None:
        """SQL prepare the dataset synchronously"""
        computation = t.cast(
            BaseComputation[t.AsyncIterator[pa.RecordBatch]],
            self.sql_computation(),
        )
        await computation.complete_task(dataset)

    async def async_sql_prepare_parents(self, dataset: st.Dataset) -> None:
        """SQL prepare all the parents of a dataset: calling async_sql_prepare
        for dataset parents and async_prepare for scalar parents.
        """
        args, kwargs = dataset.parents()
        parents = list(args) + list(kwargs.values())
        coros = [
            self.async_sql_prepare(t.cast(st.Dataset, parent))
            if dataset.prototype() == sp.Dataset
            else self.async_prepare(parent)
            for parent in parents
        ]
        # here, if many parents potentially fail, we want to be sure
        # that all of them do it, and not only the first one (to
        # modify all statuses accordingly).
        # After, we only raise the first exception for the child.

        results = await asyncio.gather(*coros, return_exceptions=True)
        exceptions = [
            element for element in results if isinstance(element, Exception)
        ]
        if len(exceptions) == 0:
            return
        raise error_aggregation(exceptions)

    def sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.Iterator[pa.RecordBatch]:
        return sync_iterator(
            self.async_sql(dataset, query, dialect, batch_size)
        )

    async def async_sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        computation = t.cast(
            BaseComputation[t.AsyncIterator[pa.RecordBatch]],
            self.sql_computation(),
        )
        return await computation.task_result(
            dataset, query=query, dialect=dialect, batch_size=batch_size
        )

    def cache_scalar(self, scalar: st.Scalar) -> None:
        sync(self.async_cache_scalar(scalar=scalar))

    def size(self, dataset: st.Dataset) -> st.Size:
        return t.cast(Size, sync(self.async_size(dataset)))

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        return t.cast(Bounds, sync(self.async_bounds(dataset)))

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        return t.cast(Marginals, sync(self.async_marginals(dataset)))

    def links(self, dataset: st.Dataset) -> st.Links:
        return t.cast(Links, sync(self.async_links(dataset)))

    def foreign_keys(self, dataset: st.Dataset) -> t.Dict[st.Path, st.Path]:
        return t.cast(
            t.Dict[st.Path, st.Path], sync(self.async_foreign_keys(dataset))
        )

    def primary_keys(self, dataset: st.Dataset) -> t.List[st.Path]:
        return t.cast(t.List[st.Path], sync(self.async_primary_keys(dataset)))

    async def async_links(self, dataset: st.Dataset) -> t.Any:
        raise NotImplementedError

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        root_dir = os.path.join(
            self.parquet_dir(), "tfrecords", dataset.uuid()
        )
        schema_type = (await self.async_schema(dataset)).type()
        signature = to_internal_signature(schema_type)

        if not os.path.exists(root_dir):
            # the dataset is cached first
            os.makedirs(root_dir)

            flattener = flatten(signature)
            serializer = serialize(signature)
            i = 0
            batches_async_it = await self.async_to_arrow(
                dataset=dataset, batch_size=BATCH_SIZE
            )
            async for batch in batches_async_it:
                filename = os.path.join(root_dir, f"batch_{i}.tfrecord")
                i += 1
                await write_tf_batch(
                    filename, batch, schema_type, flattener, serializer
                )

        # reading from cache
        glob = os.path.join(root_dir, "*.tfrecord")
        filenames = tf.data.Dataset.list_files(glob, shuffle=False)
        deserializer = deserialize(signature)
        nester = nest(signature)
        return tf.data.TFRecordDataset(filenames).map(deserializer).map(nester)

    def mock_value(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> t.Any:
        """Compute the mock value of an external transform applied on
        Dataspecs.
        """
        assert transform.is_external()
        mock_args = [arg.variant(st.ConstraintKind.MOCK) for arg in arguments]
        named_mock_args = {
            name: arg.variant(st.ConstraintKind.MOCK)
            for name, arg in named_arguments.items()
        }

        if any([ds is None for ds in mock_args]) or any(
            [ds is None for ds in named_mock_args.values()]
        ):
            raise ValueError(
                "Cannot infer the output type of the external "
                "transform because one of the parents has a None MOCK."
            )

        typed_mock_args = [t.cast(st.DataSpec, ds) for ds in mock_args]
        typed_named_mock_args = {
            name: t.cast(st.DataSpec, ds)
            for name, ds in named_mock_args.items()
        }

        try:
            mock_value = sync(
                async_compute_external_value(
                    transform, *typed_mock_args, **typed_named_mock_args
                )
            )
        except Exception as e:
            raise e

        return mock_value


async def write_tf_batch(
    filename: str,
    batch: pa.RecordBatch,
    schema_type: st.Type,
    flattener: t.Callable,
    serializer: t.Callable,
) -> None:
    with tf.io.TFRecordWriter(filename) as writer:
        batch = convert_tensorflow(
            convert_record_batch(record_batch=batch, _type=schema_type),
            schema_type,
        )
        batch = tf.data.Dataset.from_tensor_slices(batch).map(flattener)
        for row in batch:
            as_bytes = serializer(row)
            writer.write(as_bytes)


T = t.TypeVar("T")


class BaseComputation(Computation[T]):
    """General class that implements some
    methods of the protocol shared by all task
    computations"""

    task_name = ''

    def __init__(self, manager: Manager):
        self._manager = manager

    def manager(self) -> Manager:
        return self._manager

    def status(self, dataspec: st.DataSpec) -> t.Optional[st.Status]:
        return self.manager().status(dataspec, self.task_name)

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        """Launch the task computation.

        Returns an optional awaitable that can be used in async functions to
        wait for the task to complete. This can be useful if some managers have
        a more efficient way than statuses to await for the result.
        """
        raise NotImplementedError

    async def task_result(self, dataspec: st.DataSpec, **kwargs: t.Any) -> T:
        """Return the task result.

        This is the main entry point from outide the computation. The call to
        `complete_task` will launch the task if it does not exist and wait for
        it to be ready.

        Here, we assert that the task is ready and then get the result in a
        try/catch block.
        """
        status = await self.complete_task(dataspec=dataspec)
        stage = status.task(self.task_name)
        assert stage
        assert stage.ready()
        try:
            return await self.read_ready_result(
                dataspec, stage.properties(), **kwargs
            )
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=dataspec,
                manager=self.manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=dataspec,
                manager=self.manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))

    async def read_ready_result(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> T:
        """Return the task result by reading cache or computing the value."""
        raise NotImplementedError

    async def complete_task(self, dataspec: st.DataSpec) -> st.Status:
        """Poll the last status for the given task and if no status
        is available either performs the computation or delegates it
        to another manager. Then keeps polling until either the task
        is completed or an error occurs."""

        manager_status = stt.last_status(
            dataspec=dataspec, manager=self.manager(), task=self.task_name
        )

        if manager_status is None:
            task = self.launch_task(dataspec=dataspec)
            if task is not None:
                # NB: an exception raised in an asyncio task will be reraised
                # in the awaiting code
                await task
            # In the other cases, the complete_task will be reentered with
            # the pending status, resulting in a polling process
            return await self.complete_task(dataspec)

        else:
            last_task = t.cast(st.Stage, manager_status.task(self.task_name))
            if last_task.ready():
                return manager_status
            elif last_task.pending():
                return await self.pending(dataspec)
            elif last_task.processing():
                return await self.processing(dataspec)
            elif last_task.error():
                return await self.error(dataspec)
            else:
                raise ValueError(f"Inconsistent status {manager_status}")

    async def pending(self, dataspec: st.DataSpec) -> st.Status:
        """The behaviour depends on the manager"""
        raise NotImplementedError

    async def processing(self, dataspec: st.DataSpec) -> st.Status:
        """The behaviour depends on the manager"""
        raise NotImplementedError

    async def error(
        self,
        dataspec: st.DataSpec,
    ) -> st.Status:
        """The DataSpec already has an Error status.
        In this case, we clear the statuses so that the
        task can be relaunched in the future.
        """
        status = self.status(dataspec)
        assert status
        stage = status.task(self.task_name)
        assert stage
        should_clear = status_error_policy(stage=stage)
        if should_clear:
            stt.clear_task(dataspec=dataspec, task=self.task_name)
            return await self.complete_task(dataspec=dataspec)
        raise stt.DataSpecErrorStatus(
            (
                stage.properties()["relaunch"] == str(True),
                stage.properties()["message"],
            )
        )

    async def wait_for_computation(
        self,
        dataspec: st.DataSpec,
        current_stage: str,
        timeout: int = 300,
        max_delay: int = 10,
    ) -> st.Stage:
        """Utility to wait for the availability of a computation, by polling
        its status for at most `timeout` seconds, with a period of at most
        `max_delay` seconds between each attempt. Delay grows exponentially
        at first."""
        delay = min(1, timeout, max_delay)
        total_wait = 0
        while total_wait < timeout:
            status = self.status(dataspec=dataspec)
            assert status
            stage = status.task(self.task_name)
            assert stage
            if stage.stage() == current_stage:
                logger.debug(f'POLLING {self.task_name} {dataspec.uuid()}')
                await asyncio.sleep(delay)
                total_wait += delay
                delay = min(2 * delay, max_delay, timeout - total_wait)
            else:
                break
        assert stage
        return stage


class ErrorCatchingAsyncIterator:
    """Wrap an AsyncIterator and catches potential errors.

    When an error occurs, this sets the Dataspec status to error
    accordingly.
    """

    def __init__(
        self,
        ait: t.AsyncIterator,
        dataspec: st.DataSpec,
        computation: BaseComputation,
    ):
        self.ait = ait
        self.computation = computation
        self.dataspec = dataspec
        self.agen: t.Optional[t.AsyncIterator] = None

    def __aiter__(self) -> t.AsyncIterator:
        return self.ait

    async def __anext__(self) -> t.Any:
        try:
            batch = await self.ait.__anext__()
        except StopAsyncIteration:
            raise
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=self.dataspec,
                manager=self.computation.manager(),
                task=self.computation.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=self.dataspec,
                manager=self.computation.manager(),
                task=self.computation.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))

        else:
            return batch


def status_error_policy(stage: st.Stage) -> bool:
    """This methods returns whether a given error message
    should be reset to None or not. Currently, the quick
    shortcut is to check whether the error message starts
    with a TimeoutError, in this case, we want to retry"""
    assert stage.error()
    return stage.properties().get('relaunch', 'False') == str(True)


def error_aggregation(errors: t.List[Exception]) -> Exception:
    """Takes as input a list of exceptions, the first error
    that is not a DataSpecErrorStatus or that has not a relaunch.
    If they are all DataSpecErrorStatuses with relaunch, returns
    the first"""

    for error in errors:
        if isinstance(error, stt.DataSpecErrorStatus) and error.relaunch:
            continue
        else:
            return error

    return errors[0]



from __future__ import annotations

from typing import Any

try:
    import xgboost  # type: ignore[import]
except ModuleNotFoundError:
    pass  # error message in typing.py


async def xgb_fit(
    model: xgboost.XGBClassifier, *args: Any, **kwargs: Any
) -> xgboost.XGBClassifier:
    fitted_model = model.fit(*args, **kwargs)
    return fitted_model


async def xgb_predict(
    model: xgboost.XGBClassifier,
    *args: Any,
    **kwargs: Any,
) -> Any:
    y_pred = model.predict(*args, **kwargs)
    return y_pred


async def xgb_predict_proba(
    model: xgboost.XGBClassifier,
    *args: Any,
    **kwargs: Any,
) -> Any:
    y_pred = model.predict_proba(*args, **kwargs)
    return y_pred



from __future__ import annotations

from typing import (
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Protocol,
    Tuple,
    runtime_checkable,
)
import typing as t

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # Warning is displayed by typing.py

import warnings

from sarus_data_spec.storage.typing import HasStorage
import sarus_data_spec.dataspec_rewriter.typing as sdrt
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa

    sa_engine = sa.engine.Engine
except ModuleNotFoundError:
    warnings.warn("Sqlalchemy not installed, cannot send sql queries")
    sa_engine = t.Any  # type: ignore


@runtime_checkable
class Manager(st.Referrable[sp.Manager], HasStorage, Protocol):
    """Provide the dataset functionalities"""

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.Iterator[pa.RecordBatch]:
        """Synchronous method based on async_to_arrow
        that returns an iterator of arrow batches
        for the input dataset"""
        ...

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> AsyncIterator[pa.RecordBatch]:
        """Asynchronous method. It orchestrates how
        the iterator is obtained: it can either be delegated
        via arrow_task and the result polled, or computed directly
        it via the op"""
        ...

    def schema(self, dataset: st.Dataset) -> st.Schema:
        """Synchronous method that returns the schema of a
        dataspec. Based on the asynchronous version"""
        ...

    async def async_schema(self, dataset: st.Dataset) -> st.Schema:
        """Asynchronous method that returns the schema of a
        dataspec. The computation can be either delegated to
        another manager via schema_task and the result polled
        or executed directly via async_schema_ops"""
        ...

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        """Synchronous method that returns the value of a
        scalar. Based on the asynchronous version"""
        ...

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        """Asynchronous method that returns the value of a
        scalar. The computation can be either delegated to
        another manager via value_task and the result polled
        or executed directly via async_value_ops"""
        ...

    def prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready."""
        ...

    async def async_prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready asynchronously."""
        ...

    async def async_prepare_parents(self, dataspec: st.DataSpec) -> None:
        """Prepare all the parents of a Dataspec."""
        ...

    def sql_prepare(self, dataset: st.Dataset) -> None:
        """Make sure a dataset is sql ready"""
        ...

    async def async_sql_prepare(self, dataset: st.Dataset) -> None:
        """Make sure a dataset is sql ready asynchronously."""
        ...

    async def async_sql_prepare_parents(self, dataset: st.Dataset) -> None:
        """SQL prepare all the parents of a dataset. It should sql_prepare
        dataset parents and prepare Scalars parents.
        """
        ...

    def cache_scalar(self, scalar: st.Scalar) -> None:
        """Synchronous scalar caching"""
        ...

    async def async_cache_scalar(self, scalar: st.Scalar) -> None:
        """Asynchronous scalar caching"""
        ...

    def to_parquet(self, dataset: st.Dataset) -> None:
        """Synchronous parquet caching"""
        ...

    async def async_to_parquet(self, dataset: st.Dataset) -> None:
        """Asynchronous parquet caching"""
        ...

    def parquet_dir(self) -> str:
        ...

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        ...

    async def async_marginals(self, dataset: st.Dataset) -> st.Marginals:
        ...

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        ...

    async def async_bounds(self, dataset: st.Dataset) -> st.Bounds:
        ...

    def size(self, dataset: st.Dataset) -> st.Size:
        ...

    async def async_size(self, dataset: st.Dataset) -> st.Size:
        ...

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        ...

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        ...

    def to_sql(self, dataset: st.Dataset) -> None:
        ...

    async def async_to_sql(self, dataset: st.Dataset) -> None:
        ...

    def status(
        self, dataspec: st.DataSpec, task_name: str
    ) -> t.Optional[st.Status]:
        ...

    def dataspec_rewriter(self) -> sdrt.DataspecRewriter:
        ...

    def dataspec_validator(self) -> sdvt.DataspecValidator:
        ...

    def is_remote(self, dataspec: st.DataSpec) -> bool:
        """Is the dataspec a remotely defined dataset."""
        ...

    def infer_output_type(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> Tuple[str, Callable[[st.DataSpec], None]]:
        ...

    def foreign_keys(self, dataset: st.Dataset) -> Dict[st.Path, st.Path]:
        ...

    async def async_foreign_keys(
        self, dataset: st.Dataset
    ) -> Dict[st.Path, st.Path]:
        ...

    async def async_primary_keys(self, dataset: st.Dataset) -> List[st.Path]:
        ...

    def primary_keys(self, dataset: st.Dataset) -> List[st.Path]:
        ...

    def sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> Iterator[pa.RecordBatch]:
        ...

    async def async_sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> AsyncIterator[pa.RecordBatch]:
        ...

    async def execute_sql_query(
        self,
        dataset: st.Dataset,
        caching_properties: t.Mapping[str, str],
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    async def async_sql_op(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Mapping[t.Union[str, t.Tuple[str, ...]], str]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    def is_big_data(self, dataset: st.Dataset) -> bool:
        ...

    def is_cached(self, dataspec: st.DataSpec) -> bool:
        """Returns whether a dataspec should be cached
        or not"""
        ...

    def is_pushed_to_sql(self, dataspec: st.DataSpec) -> bool:
        """Returns whether a dataspec should be pushed to sql
        or not"""
        ...

    def attribute(
        self, name: str, dataspec: st.DataSpec
    ) -> t.Optional[st.Attribute]:
        ...

    def attributes(
        self, name: str, dataspec: st.DataSpec
    ) -> t.List[st.Attribute]:
        ...

    def links(self, dataset: st.Dataset) -> st.Links:
        ...

    async def async_links(self, dataset: st.Dataset) -> st.Links:
        ...

    def sql_pushing_schema_prefix(self, dataset: st.Dataset) -> str:
        ...

    def engine(self, uri: str) -> sa_engine:
        ...

    def mock_value(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> t.Any:
        """Compute the mock value of an external transform applied on
        Dataspecs.
        """


@runtime_checkable
class HasManager(Protocol):
    """Has a manager."""

    def manager(self) -> Manager:
        """Return a manager (usually a singleton)."""
        ...


T = t.TypeVar("T", covariant=True)


class Computation(t.Protocol[T]):
    """Protocol for classes that perform tasks computations.
    It sets how computations are scheduled and launched
    depending on statuses. A computation is mainly defined by two methods:
     - launch : a method that does not return a value but
     that only has side effects, changing either the storage or the cache
    and that updates statuses during the process
    - result: a method that allows to get the value of the computation
    either by reading the cache/storage or via the ops.

    Furthermore, a computation has a method to monitor task completion.
    """

    task_name: str = ''

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        """This methods launches a task in the background
        but returns immediately without waiting for the
        result. It updates the statuses during its process."""
        ...

    async def task_result(self, dataspec: st.DataSpec, **kwargs: t.Any) -> T:
        """Returns the value for the given computed task. It either
        retrieves it from the cache or computes it via the ops."""
        ...

    async def complete_task(self, dataspec: st.DataSpec) -> st.Status:
        """Monitors a task: it launches it if there is no status
        and then polls until it is ready/error"""
        ...


class DelegatedComputation(Computation, t.Protocol[T]):
    def delegate_manager_status(
        self, dataspec: st.DataSpec
    ) -> Optional[st.Status]:
        """While some managers are eager and always execute tasks,
        others are lazy and delegate them. This is an interface
        for a manager to get the status of the eager manager
        that is executing the task"""
        ...


@t.runtime_checkable
class ExternalOpImplementation(t.Protocol):
    """External Op implementation class.

    The `allowed_pep_args` is a list of combinations of arguments' names which
    are managed by the Op. The result of the Op will be PEP only if the set of
    PEP arguments passed to the Op are in this list.

    For instance, if we have an op that takes 3 arguments `a`, `b` and `c` and
    the `allowed_pep_args` are [{'a'}, {'b'}, {'a','b'}] then the following
    combinations will yield a PEP output:
        - `a` is a PEP dataspec, `b` and `c` are either not dataspecs or public
          dataspecs
        - `b` is a PEP dataspec, `a` and `c` are either not dataspecs or public
          dataspecs
        - `a` and `b` are PEP dataspecs, `c` is either not a dataspec or a
          public dataspec
    """

    transform_id: str
    allowed_pep_args: t.List[t.Set[str]]

    def dataspec(self) -> st.DataSpec:
        ...

    def dp_equivalent(self) -> t.Optional[ExternalOpImplementation]:
        ...

    @staticmethod
    async def call(*args: t.Any, **kwargs: t.Any) -> t.Any:
        """Compute the external op output value.

        DP implementation take additional arguments:
            - `seed` an integer used to parametrize rangom number generators
            - `budget` the privacy budget that can be spend in the op
            - `pe` the protected entity information of each row
        """

    def pep_kind(self, *args: t.Any, **kwargs: t.Any) -> sdvt.PEPKind:
        """Return the PEP properties of the transform.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, it is not PEP if we are aggregating
        the rows (axis=0) and it is PEP if we are aggregating the columns
        (axis=1).
        """

    def is_dp(self, *args: t.Any, **kwargs: t.Any) -> bool:
        """Return True if the DP transform is compatible with the arguments.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, if we are aggregating the rows
        (axis=0), then there might be an equivalent DP transform but if we are
        aggregating the columns there might not (axis=1).

        NB: This function should have the same signature as the call function.
        """

    async def private_queries(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""



from typing import Any

try:
    from pandas_profiling import ProfileReport
except ModuleNotFoundError:
    pass  # error message in typing.py


async def pd_profile_report(df: Any, *args: Any, **kwargs: Any) -> Any:
    return ProfileReport(df, *args, **kwargs)
